from pydantic import Field
from koil.composition import Composition
from .rath import OmeroArkRath


class OmeroArk(Composition):
    """The Mikro Composition

    This composition provides a datalayer and a omero_ark for interacting with the
    mikro api and beyond

    """

    rath: OmeroArkRath

    def _repr_html_inline_(self):
        return f"<table><td>rath</td><td>{self.rath._repr_html_inline_()}</td></tr></table>"



from dokker import mirror,  Deployment
import os
from koil.composition import Composition
from rath.links.auth import ComposedAuthLink
from rath.links.aiohttp import AIOHttpLink
from rath.links.graphql_ws import GraphQLWSLink
from omero_ark.omero_ark import OmeroArk
from omero_ark.rath import (
    OmeroArkRath,
    SplitLink,
    OmeroArkRathLinkComposition,
)
from graphql import OperationType

test_path = os.path.join(os.path.dirname(__file__), "deployments", "test")


def build_deployment() -> Deployment:
    setup = mirror(test_path)
    setup.add_health_check(
        url="http://localhost:7755/graphql", service="omero_ark", timeout=5, max_retries=10
    )
    return setup


async def token_loader():
    """ Returns a token as defined in the static_token setting for omero_ark"""
    return "demo"


def build_deployed_omero_ark() -> OmeroArk:

    y = OmeroArkRath(
        link=OmeroArkRathLinkComposition(
            auth=ComposedAuthLink(
                token_loader=token_loader,
                token_refresher=token_loader,
            ),
            split=SplitLink(
                left=AIOHttpLink(endpoint_url="http://localhost:7755/graphql"),
                right=GraphQLWSLink(ws_endpoint_url="ws://localhost:7755/graphql"),
                split=lambda o: o.node.operation != OperationType.SUBSCRIPTION,
            ),
        ),
    )

    omero_ark = OmeroArk(
        rath=y,
    )
    return omero_ark


class DeployedOmeroArk(Composition):
    """ A deployed omero_ark"""
    deployment: Deployment
    omero_ark: OmeroArk


def deployed() -> DeployedOmeroArk:
    """Create a deployed omero_ark

    A deployed omero_ark is a composition of a deployment and a omero_ark.
    This means a fully functioning omero instance will be spun up when
    the context manager is entered.

    To inspect the deployment, use the `deployment` attribute.
    To interact with the omero_ark, use the `omero_ark` attribute.


    Returns
    -------
    DeployedOmeroArk
        _description_
    """
    return DeployedOmeroArk(
        deployment=build_deployment(),
        omero_ark=build_deployed_omero_ark(),
    )



from omero_ark.rath import OmeroArkRath
from omero_ark.funcs import execute, aexecute
from typing_extensions import Literal
from typing import Optional, Tuple, List
from enum import Enum
from pydantic import Field, BaseModel
from datetime import datetime
from rath.scalars import ID


class ProjectFragmentDatasets(BaseModel):
    typename: Optional[Literal["Dataset"]] = Field(alias="__typename", exclude=True)
    id: str
    name: str
    description: str

    class Config:
        """A config class"""

        frozen = True


class ProjectFragment(BaseModel):
    typename: Optional[Literal["Project"]] = Field(alias="__typename", exclude=True)
    id: str
    name: str
    description: str
    datasets: Tuple[ProjectFragmentDatasets, ...]

    class Config:
        """A config class"""

        frozen = True


class DatasetFragmentImages(BaseModel):
    typename: Optional[Literal["Image"]] = Field(alias="__typename", exclude=True)
    id: str
    name: str
    description: str

    class Config:
        """A config class"""

        frozen = True


class DatasetFragment(BaseModel):
    typename: Optional[Literal["Dataset"]] = Field(alias="__typename", exclude=True)
    id: str
    name: str
    description: str
    images: Tuple[DatasetFragmentImages, ...]

    class Config:
        """A config class"""

        frozen = True


class ImageFragment(BaseModel):
    typename: Optional[Literal["Image"]] = Field(alias="__typename", exclude=True)
    id: str
    name: str
    description: str

    class Config:
        """A config class"""

        frozen = True


class EnsureOmeroUserMutationEnsureomerouserUser(BaseModel):
    typename: Optional[Literal["User"]] = Field(alias="__typename", exclude=True)
    id: ID
    sub: str

    class Config:
        """A config class"""

        frozen = True


class EnsureOmeroUserMutationEnsureomerouser(BaseModel):
    typename: Optional[Literal["OmeroUser"]] = Field(alias="__typename", exclude=True)
    id: ID
    omero_username: str = Field(alias="omeroUsername")
    omero_password: str = Field(alias="omeroPassword")
    user: EnsureOmeroUserMutationEnsureomerouserUser

    class Config:
        """A config class"""

        frozen = True


class EnsureOmeroUserMutation(BaseModel):
    ensure_omero_user: EnsureOmeroUserMutationEnsureomerouser = Field(
        alias="ensureOmeroUser"
    )

    class Arguments(BaseModel):
        username: str
        password: str

    class Meta:
        document = "mutation EnsureOmeroUser($username: String!, $password: String!) {\n  ensureOmeroUser(input: {username: $username, password: $password}) {\n    id\n    omeroUsername\n    omeroPassword\n    user {\n      id\n      sub\n    }\n  }\n}"


class ListProjectsQueryProjectsDatasetsImages(BaseModel):
    typename: Optional[Literal["Image"]] = Field(alias="__typename", exclude=True)
    name: str
    acquisition_date: Optional[datetime] = Field(alias="acquisitionDate")

    class Config:
        """A config class"""

        frozen = True


class ListProjectsQueryProjectsDatasets(BaseModel):
    typename: Optional[Literal["Dataset"]] = Field(alias="__typename", exclude=True)
    name: str
    description: str
    images: Tuple[ListProjectsQueryProjectsDatasetsImages, ...]

    class Config:
        """A config class"""

        frozen = True


class ListProjectsQueryProjects(BaseModel):
    typename: Optional[Literal["Project"]] = Field(alias="__typename", exclude=True)
    name: str
    description: str
    datasets: Tuple[ListProjectsQueryProjectsDatasets, ...]

    class Config:
        """A config class"""

        frozen = True


class ListProjectsQuery(BaseModel):
    projects: Tuple[ListProjectsQueryProjects, ...]

    class Arguments(BaseModel):
        pass

    class Meta:
        document = "query ListProjects {\n  projects {\n    name\n    description\n    datasets {\n      name\n      description\n      images {\n        name\n        acquisitionDate\n      }\n    }\n  }\n}"


class GetProjectQuery(BaseModel):
    project: ProjectFragment

    class Arguments(BaseModel):
        id: ID

    class Meta:
        document = "fragment Project on Project {\n  id\n  name\n  description\n  datasets {\n    id\n    name\n    description\n  }\n}\n\nquery GetProject($id: ID!) {\n  project(id: $id) {\n    ...Project\n  }\n}"


class SearchProjectsQueryOptions(BaseModel):
    typename: Optional[Literal["Project"]] = Field(alias="__typename", exclude=True)
    value: str
    label: str

    class Config:
        """A config class"""

        frozen = True


class SearchProjectsQuery(BaseModel):
    options: Tuple[SearchProjectsQueryOptions, ...]

    class Arguments(BaseModel):
        search: Optional[str] = Field(default=None)
        values: Optional[List[ID]] = Field(default=None)

    class Meta:
        document = "query SearchProjects($search: String, $values: [ID!]) {\n  options: projects(\n    filters: {search: $search, ids: $values}\n    pagination: {limit: 10}\n  ) {\n    value: id\n    label: name\n  }\n}"


class MeQueryMeOmerouser(BaseModel):
    typename: Optional[Literal["OmeroUser"]] = Field(alias="__typename", exclude=True)
    omero_password: str = Field(alias="omeroPassword")
    omero_username: str = Field(alias="omeroUsername")

    class Config:
        """A config class"""

        frozen = True


class MeQueryMe(BaseModel):
    typename: Optional[Literal["User"]] = Field(alias="__typename", exclude=True)
    omero_user: Optional[MeQueryMeOmerouser] = Field(alias="omeroUser")

    class Config:
        """A config class"""

        frozen = True


class MeQuery(BaseModel):
    me: MeQueryMe

    class Arguments(BaseModel):
        pass

    class Meta:
        document = "query Me {\n  me {\n    omeroUser {\n      omeroPassword\n      omeroUsername\n    }\n  }\n}"


class ListDatasetsQuery(BaseModel):
    datasets: DatasetFragment

    class Arguments(BaseModel):
        pass

    class Meta:
        document = "fragment Dataset on Dataset {\n  id\n  name\n  description\n  images {\n    id\n    name\n    description\n  }\n}\n\nquery ListDatasets {\n  datasets {\n    ...Dataset\n  }\n}"


class GetDatasetQuery(BaseModel):
    dataset: DatasetFragment

    class Arguments(BaseModel):
        id: ID

    class Meta:
        document = "fragment Dataset on Dataset {\n  id\n  name\n  description\n  images {\n    id\n    name\n    description\n  }\n}\n\nquery GetDataset($id: ID!) {\n  dataset(id: $id) {\n    ...Dataset\n  }\n}"


class SearchDatasetsQueryOptions(BaseModel):
    typename: Optional[Literal["Dataset"]] = Field(alias="__typename", exclude=True)
    value: str
    label: str

    class Config:
        """A config class"""

        frozen = True


class SearchDatasetsQuery(BaseModel):
    options: SearchDatasetsQueryOptions

    class Arguments(BaseModel):
        search: Optional[str] = Field(default=None)
        values: Optional[List[ID]] = Field(default=None)

    class Meta:
        document = "query SearchDatasets($search: String, $values: [ID!]) {\n  options: datasets(\n    filters: {search: $search, ids: $values}\n    pagination: {limit: 10}\n  ) {\n    value: id\n    label: name\n  }\n}"


class ListImagesQueryImages(BaseModel):
    typename: Optional[Literal["Image"]] = Field(alias="__typename", exclude=True)
    name: str
    description: str

    class Config:
        """A config class"""

        frozen = True


class ListImagesQuery(BaseModel):
    images: ListImagesQueryImages

    class Arguments(BaseModel):
        pass

    class Meta:
        document = "query ListImages {\n  images {\n    name\n    description\n  }\n}"


class GetImageQuery(BaseModel):
    image: ImageFragment

    class Arguments(BaseModel):
        id: ID

    class Meta:
        document = "fragment Image on Image {\n  id\n  name\n  description\n}\n\nquery GetImage($id: ID!) {\n  image(id: $id) {\n    ...Image\n  }\n}"


class SearchImagesQueryOptions(BaseModel):
    typename: Optional[Literal["Image"]] = Field(alias="__typename", exclude=True)
    value: str
    label: str

    class Config:
        """A config class"""

        frozen = True


class SearchImagesQuery(BaseModel):
    options: SearchImagesQueryOptions

    class Arguments(BaseModel):
        search: Optional[str] = Field(default=None)
        values: Optional[List[ID]] = Field(default=None)

    class Meta:
        document = "query SearchImages($search: String, $values: [ID!]) {\n  options: images(\n    filters: {search: $search, ids: $values}\n    pagination: {limit: 10}\n  ) {\n    value: id\n    label: name\n  }\n}"


async def aensure_omero_user(
    username: str, password: str, rath: Optional[OmeroArkRath] = None
) -> EnsureOmeroUserMutationEnsureomerouser:
    """EnsureOmeroUser



    Arguments:
        username (str): username
        password (str): password
        rath (omero_ark.rath.OmeroArkRath, optional): The omero_ark rath client

    Returns:
        EnsureOmeroUserMutationEnsureomerouser"""
    return (
        await aexecute(
            EnsureOmeroUserMutation,
            {"username": username, "password": password},
            rath=rath,
        )
    ).ensure_omero_user


def ensure_omero_user(
    username: str, password: str, rath: Optional[OmeroArkRath] = None
) -> EnsureOmeroUserMutationEnsureomerouser:
    """EnsureOmeroUser



    Arguments:
        username (str): username
        password (str): password
        rath (omero_ark.rath.OmeroArkRath, optional): The omero_ark rath client

    Returns:
        EnsureOmeroUserMutationEnsureomerouser"""
    return execute(
        EnsureOmeroUserMutation, {"username": username, "password": password}, rath=rath
    ).ensure_omero_user


async def alist_projects(
    rath: Optional[OmeroArkRath] = None,
) -> List[ListProjectsQueryProjects]:
    """ListProjects



    Arguments:
        rath (omero_ark.rath.OmeroArkRath, optional): The omero_ark rath client

    Returns:
        List[ListProjectsQueryProjects]"""
    return (await aexecute(ListProjectsQuery, {}, rath=rath)).projects


def list_projects(
    rath: Optional[OmeroArkRath] = None,
) -> List[ListProjectsQueryProjects]:
    """ListProjects



    Arguments:
        rath (omero_ark.rath.OmeroArkRath, optional): The omero_ark rath client

    Returns:
        List[ListProjectsQueryProjects]"""
    return execute(ListProjectsQuery, {}, rath=rath).projects


async def aget_project(id: ID, rath: Optional[OmeroArkRath] = None) -> ProjectFragment:
    """GetProject



    Arguments:
        id (ID): id
        rath (omero_ark.rath.OmeroArkRath, optional): The omero_ark rath client

    Returns:
        ProjectFragment"""
    return (await aexecute(GetProjectQuery, {"id": id}, rath=rath)).project


def get_project(id: ID, rath: Optional[OmeroArkRath] = None) -> ProjectFragment:
    """GetProject



    Arguments:
        id (ID): id
        rath (omero_ark.rath.OmeroArkRath, optional): The omero_ark rath client

    Returns:
        ProjectFragment"""
    return execute(GetProjectQuery, {"id": id}, rath=rath).project


async def asearch_projects(
    search: Optional[str] = None,
    values: Optional[List[ID]] = None,
    rath: Optional[OmeroArkRath] = None,
) -> List[SearchProjectsQueryOptions]:
    """SearchProjects



    Arguments:
        search (Optional[str], optional): search.
        values (Optional[List[ID]], optional): values.
        rath (omero_ark.rath.OmeroArkRath, optional): The omero_ark rath client

    Returns:
        List[SearchProjectsQueryProjects]"""
    return (
        await aexecute(
            SearchProjectsQuery, {"search": search, "values": values}, rath=rath
        )
    ).options


def search_projects(
    search: Optional[str] = None,
    values: Optional[List[ID]] = None,
    rath: Optional[OmeroArkRath] = None,
) -> List[SearchProjectsQueryOptions]:
    """SearchProjects



    Arguments:
        search (Optional[str], optional): search.
        values (Optional[List[ID]], optional): values.
        rath (omero_ark.rath.OmeroArkRath, optional): The omero_ark rath client

    Returns:
        List[SearchProjectsQueryProjects]"""
    return execute(
        SearchProjectsQuery, {"search": search, "values": values}, rath=rath
    ).options


async def ame(rath: Optional[OmeroArkRath] = None) -> MeQueryMe:
    """Me



    Arguments:
        rath (omero_ark.rath.OmeroArkRath, optional): The omero_ark rath client

    Returns:
        MeQueryMe"""
    return (await aexecute(MeQuery, {}, rath=rath)).me


def me(rath: Optional[OmeroArkRath] = None) -> MeQueryMe:
    """Me



    Arguments:
        rath (omero_ark.rath.OmeroArkRath, optional): The omero_ark rath client

    Returns:
        MeQueryMe"""
    return execute(MeQuery, {}, rath=rath).me


async def alist_datasets(rath: Optional[OmeroArkRath] = None) -> DatasetFragment:
    """ListDatasets



    Arguments:
        rath (omero_ark.rath.OmeroArkRath, optional): The omero_ark rath client

    Returns:
        DatasetFragment"""
    return (await aexecute(ListDatasetsQuery, {}, rath=rath)).datasets


def list_datasets(rath: Optional[OmeroArkRath] = None) -> DatasetFragment:
    """ListDatasets



    Arguments:
        rath (omero_ark.rath.OmeroArkRath, optional): The omero_ark rath client

    Returns:
        DatasetFragment"""
    return execute(ListDatasetsQuery, {}, rath=rath).datasets


async def aget_dataset(id: ID, rath: Optional[OmeroArkRath] = None) -> DatasetFragment:
    """GetDataset



    Arguments:
        id (ID): id
        rath (omero_ark.rath.OmeroArkRath, optional): The omero_ark rath client

    Returns:
        DatasetFragment"""
    return (await aexecute(GetDatasetQuery, {"id": id}, rath=rath)).dataset


def get_dataset(id: ID, rath: Optional[OmeroArkRath] = None) -> DatasetFragment:
    """GetDataset



    Arguments:
        id (ID): id
        rath (omero_ark.rath.OmeroArkRath, optional): The omero_ark rath client

    Returns:
        DatasetFragment"""
    return execute(GetDatasetQuery, {"id": id}, rath=rath).dataset


async def asearch_datasets(
    search: Optional[str] = None,
    values: Optional[List[ID]] = None,
    rath: Optional[OmeroArkRath] = None,
) -> SearchDatasetsQueryOptions:
    """SearchDatasets



    Arguments:
        search (Optional[str], optional): search.
        values (Optional[List[ID]], optional): values.
        rath (omero_ark.rath.OmeroArkRath, optional): The omero_ark rath client

    Returns:
        SearchDatasetsQueryDatasets"""
    return (
        await aexecute(
            SearchDatasetsQuery, {"search": search, "values": values}, rath=rath
        )
    ).options


def search_datasets(
    search: Optional[str] = None,
    values: Optional[List[ID]] = None,
    rath: Optional[OmeroArkRath] = None,
) -> SearchDatasetsQueryOptions:
    """SearchDatasets



    Arguments:
        search (Optional[str], optional): search.
        values (Optional[List[ID]], optional): values.
        rath (omero_ark.rath.OmeroArkRath, optional): The omero_ark rath client

    Returns:
        SearchDatasetsQueryDatasets"""
    return execute(
        SearchDatasetsQuery, {"search": search, "values": values}, rath=rath
    ).options


async def alist_images(rath: Optional[OmeroArkRath] = None) -> ListImagesQueryImages:
    """ListImages



    Arguments:
        rath (omero_ark.rath.OmeroArkRath, optional): The omero_ark rath client

    Returns:
        ListImagesQueryImages"""
    return (await aexecute(ListImagesQuery, {}, rath=rath)).images


def list_images(rath: Optional[OmeroArkRath] = None) -> ListImagesQueryImages:
    """ListImages



    Arguments:
        rath (omero_ark.rath.OmeroArkRath, optional): The omero_ark rath client

    Returns:
        ListImagesQueryImages"""
    return execute(ListImagesQuery, {}, rath=rath).images


async def aget_image(id: ID, rath: Optional[OmeroArkRath] = None) -> ImageFragment:
    """GetImage



    Arguments:
        id (ID): id
        rath (omero_ark.rath.OmeroArkRath, optional): The omero_ark rath client

    Returns:
        ImageFragment"""
    return (await aexecute(GetImageQuery, {"id": id}, rath=rath)).image


def get_image(id: ID, rath: Optional[OmeroArkRath] = None) -> ImageFragment:
    """GetImage



    Arguments:
        id (ID): id
        rath (omero_ark.rath.OmeroArkRath, optional): The omero_ark rath client

    Returns:
        ImageFragment"""
    return execute(GetImageQuery, {"id": id}, rath=rath).image


async def asearch_images(
    search: Optional[str] = None,
    values: Optional[List[ID]] = None,
    rath: Optional[OmeroArkRath] = None,
) -> SearchImagesQueryOptions:
    """SearchImages



    Arguments:
        search (Optional[str], optional): search.
        values (Optional[List[ID]], optional): values.
        rath (omero_ark.rath.OmeroArkRath, optional): The omero_ark rath client

    Returns:
        SearchImagesQueryImages"""
    return (
        await aexecute(
            SearchImagesQuery, {"search": search, "values": values}, rath=rath
        )
    ).options


def search_images(
    search: Optional[str] = None,
    values: Optional[List[ID]] = None,
    rath: Optional[OmeroArkRath] = None,
) -> SearchImagesQueryOptions:
    """SearchImages



    Arguments:
        search (Optional[str], optional): search.
        values (Optional[List[ID]], optional): values.
        rath (omero_ark.rath.OmeroArkRath, optional): The omero_ark rath client

    Returns:
        SearchImagesQueryImages"""
    return execute(
        SearchImagesQuery, {"search": search, "values": values}, rath=rath
    ).options



""" This module provides helpers for the mikro rath api
they are wrapped functions for the turms generated api"""
from .rath import OmeroArkRath, current_omero_ark_rath
from koil.helpers import unkoil, unkoil_gen


async def aexecute(
    operation,
    variables,
    rath: OmeroArkRath = None,
):
    rath = rath or current_omero_ark_rath.get()

    x = await rath.aquery(
        operation.Meta.document,
        operation.Arguments(**variables).dict(by_alias=True),
    )
    return operation(**x.data)


def execute(
    operation,
    variables,
    rath: OmeroArkRath = None,
):
    return unkoil(aexecute, operation, variables, rath=rath)


def subscribe(operation, variables, rath: OmeroArkRath = None):
    return unkoil_gen(asubscribe, operation, variables, rath=rath)


async def asubscribe(operation, variables, rath: OmeroArkRath = None):
    rath = rath or current_omero_ark_rath.get()
    async for event in rath.asubscribe(
        operation.Meta.document, operation.Arguments(**variables)
    ):
        yield operation(**event.data)



"""
Traits for omero_ark


Traits are mixins that are added to every graphql type that exists on the mikro schema.
We use them to add functionality to the graphql types that extend from the base type.

Every GraphQL Model on Mikro gets a identifier and shrinking methods to ensure the compatibliity
with arkitekt. This is done by adding the identifier and the shrinking methods to the graphql type.
If you want to add your own traits to the graphql type, you can do so by adding them in the graphql
.config.yaml file.

"""

from typing import Awaitable, List, TypeVar, Tuple, Protocol, Optional
import numpy as np
from pydantic import BaseModel
import xarray as xr
import pandas as pd
from rath.links.shrink import ShrinkByID
from typing import TYPE_CHECKING
from rath.scalars import ID
from typing import Any
import pyarrow.parquet as pq

if TYPE_CHECKING:
    pass


class Image(BaseModel):
    """Image Trait

    This trait is added to every graphql type that extends from the Image type.
    It adds the identifier and the shrinking methods to the graphql type.

    """

    async def adownlaod(self, filepath: str) -> Awaitable[xr.DataArray]:
        """The Data of the Representation as an xr.DataArray. Accessible from asyncio.

        Returns:
            xr.DataArray: The associated object.

        Raises:
            AssertionError: If the representation has no store attribute quries
        """
        import aiohttp
        import aiofiles

        async with aiohttp.ClientSession() as session:
            url = self.file
            async with session.get(url) as resp:
                if resp.status == 200:
                    f = await aiofiles.open("/some/file.img", mode="wb")
                    await f.write(await resp.read())
                    await f.close()



""" Strucutre Registration

This module is autoimported by arkitekt. It registers the default structure types with the arkitekt
structure-registry so that they can be used in the arkitekt app without having to import them.

You can of course overwrite this in your app if you need to expand to a more complex query.

"""
import logging

logger = logging.getLogger(__name__)


try:
    import rekuest
    from rekuest.structures.default import (
        get_default_structure_registry,
        Scope,
        id_shrink,
    )
    from rekuest.widgets import SearchWidget
    from omero_ark.api.schema import (
        ProjectFragment,
        aget_project,
        SearchProjectsQuery,
        aget_image,
        aget_dataset,
        ImageFragment,
        SearchImagesQuery,
        DatasetFragment,
        SearchDatasetsQuery,
    )

    structure_reg = get_default_structure_registry()
    structure_reg.register_as_structure(
        ProjectFragment,
        identifier="@omero-ark/project",
        aexpand=aget_project,
        ashrink=id_shrink,
        scope=Scope.GLOBAL,
        default_widget=SearchWidget(
            query=SearchProjectsQuery.Meta.document, ward="omero-ark"
        ),
    )
    structure_reg.register_as_structure(
        ImageFragment,
        identifier="@omero-ark/image",
        aexpand=aget_image,
        ashrink=id_shrink,
        scope=Scope.GLOBAL,
        default_widget=SearchWidget(
            query=SearchImagesQuery.Meta.document, ward="omero-ark"
        ),
    )
    structure_reg.register_as_structure(
        DatasetFragment,
        identifier="@omero-ark/dataset",
        aexpand=aget_dataset,
        ashrink=id_shrink,
        scope=Scope.GLOBAL,
        default_widget=SearchWidget(
            query=SearchDatasetsQuery.Meta.document, ward="omero-ark"
        ),
    )


except ImportError:
    pass
